import logging
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from bokeh.plotting import figure, save
from bokeh.models import HoverTool, ColumnDataSource
from bokeh.io import output_file

# Add logging 
logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

# Read in tsv containing p-values and LFCs
logger.info("Loading p-values and LFCs...")
data = pd.read_table(snakemake.input[0], sep="\t", index_col=0).squeeze()
pval_col = snakemake.params["pval_col"]
diff_ab_col = snakemake.params["diff_ab_col"]

# Take negative log of pval_col column and add as new column
data['negative_log_pval'] = -np.log10(data[pval_col])

# Create a ColumnDataSource object
source_blue = ColumnDataSource(data=dict(
    diff_ab_col=data[diff_ab_col],
    pval=data['negative_log_pval'],
    feature=data.index
))

source_red = ColumnDataSource(data=dict(
    diff_ab_col=data[diff_ab_col],
    pval=data['negative_log_pval'],
    feature=data.index
))

# Create a figure
tool_name = snakemake.wildcards["tool"]
p = figure(title=f"{tool_name} Volcano Plot", x_axis_label='coefficient', y_axis_label='-log10(pval_col)')

p.title.text_font_size = '20pt'

# Define the desired y-axis range
y_min = data['negative_log_pval'].min() - 1  # Minimum value
y_max = data['negative_log_pval'].max() + 1  # Maximum value

# Set the y-axis range
p.y_range.start = y_min
p.y_range.end = y_max

# Define the desired x-axis range
x_min = data[diff_ab_col].min() - 1  # Minimum value
x_max = data[diff_ab_col].max() + 1  # Maximum value

# Set the y-axis range
p.x_range.start = x_min
p.x_range.end = x_max

# # Add a vertical line at x = -1
p.line(x=[-1, -1], y=[y_min, y_max], line_color='gray', line_width=1, line_dash='dashed')
p.line(x=[1, 1], y=[y_min, y_max], line_color='gray', line_width=1, line_dash='dashed')
p.line(x=[x_min,x_max], y=[-np.log10(0.05), -np.log10(0.05)], line_color='gray', line_width=1, line_dash='dashed')

# Add hover tool
hover = HoverTool(
    tooltips=[
        ('Feature', '@feature'),
        ('coefficient', '@diff_ab_col'),
        ('-log10(pval_col)', '@pval'),
    ]
)
p.add_tools(hover)

# Select points where y >= 1 and x >= -log10(0.05)
pos_sig_diff = (data[diff_ab_col] >= 1) & (data['negative_log_pval'] >= -np.log10(0.05))
# Select points where y <= -1 and x >= -log10(0.05)
neg_sig_diff = (data[diff_ab_col] <= -1) & (data['negative_log_pval'] >= -np.log10(0.05))

# Apply selection to the data source
source_blue.selected.indices = np.where(neg_sig_diff)[0]
# Apply selection to the data source for negative significant differences (blue)
p.circle('diff_ab_col', 'pval', size=5, fill_alpha=0.6, nonselection_line_color='black', nonselection_fill_color='black', source=source_blue, selection_color='blue')

# Apply selection to the data source
source_red.selected.indices = np.where(pos_sig_diff)[0]
# Apply selection to the data source for positive significant differences (red)
p.circle('diff_ab_col', 'pval', size=5, fill_alpha=0.6, nonselection_line_color='black', nonselection_fill_color='black', source=source_red, selection_color='red')


output_file(snakemake.output[0])
save(p)

logger.info(f"Saved volocano plot to {snakemake.output[0]}")



import logging
import pandas as pd

logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

logger.info("Loading p-values...")
pvalue_files = [
    pd.read_table(x, sep="\t", index_col=0) for x in snakemake.input
]
concat_pvalues = pd.concat(pvalue_files, axis=1)

concat_pvalues.to_csv(snakemake.output[0], sep="\t", index=True)
logger.info(f"Saved to {snakemake.output[0]}")



from collections import defaultdict
import logging

import joblib
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

tools = snakemake.config["tools"] + ["pca_pc1"]
palette = dict(zip(
    tools, sns.color_palette("colorblind", len(tools))
))

pctile = snakemake.wildcards["pctile"]

models = defaultdict(dict)
for tool, model_loc in zip(tools, snakemake.input):
    logger.info(f"Loading {model_loc}...")
    models[tool] = joblib.load(model_loc)

fig, ax = plt.subplots(1, 1)
ax.set_aspect("equal")

best_tool = None
best_auc = 0
leg_lines = []
for tool in tools:
    this_model_data = models[tool]
    mean_fprs = this_model_data["fprs"]
    mean_tprs = np.mean(this_model_data["tprs"], axis=0)

    roc_aucs = this_model_data["roc_aucs"]
    mean_auc = np.mean(roc_aucs)
    std_auc = np.std(roc_aucs)

    if mean_auc > best_auc:
        best_auc = mean_auc
        best_tool = tool

    color = palette[tool]
    line = Line2D([0], [0], color=color, lw=2,
                  label=f"{tool} ({mean_auc:.2f} $\pm$ {std_auc:.2f})")
    leg_lines.append(line)

    ax.plot(mean_fprs, mean_tprs, lw=2, color=color)
    logger.info(f"{tool} Mean AUC = {mean_auc:.2f} +- {std_auc:.2f}")

leg = ax.legend(handles=leg_lines, loc="lower right", frameon=False)
for text in leg.get_texts():
    if best_tool in text._text:
        text.set_weight("bold")

logger.info(f"Best model: {best_tool} ({best_auc:.2f})")

ax.plot([0, 1], [0, 1], linestyle="--", lw=2, color="black")
ax.set_xlabel("False Positive Rate")
ax.set_ylabel("True Positive Rate")
ax.set_title(f"{pctile}% Features")

plt.savefig(snakemake.output[0])
logger.info(f"Saved to {snakemake.output[0]}")



import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

logger.info("Loading pvalues...")
diffs_df = pd.read_table(snakemake.input[0], sep="\t", index_col=0)

corr = diffs_df.corr("kendall")
mask = np.zeros_like(corr)
mask[np.triu_indices_from(mask)] = True

fig, ax = plt.subplots(1, 1)

palette = sns.color_palette("rocket", as_cmap=True)

sns.heatmap(
    corr,
    annot=True,
    mask=mask,
    square=True,
    ax=ax,
    cmap=palette,
    vmin=0,
    vmax=1
)
ax.set_title("Kendall Rank Correlations of P-values")
plt.savefig(snakemake.output[0])
logger.info(f"Saved to {snakemake.output[0]}")



import logging
import pandas as pd

logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

logger.info("Loading differentials...")
concatenated_diff_file = pd.read_csv(snakemake.input["concatenated_differentials"], sep="\t", index_col=0)

for col in concatenated_diff_file.columns:
    new_col_name = col + " differentials"
    concatenated_diff_file.rename(columns={col: new_col_name}, inplace=True)
    
logger.info("Loading p-values...")
concatenated_pvalue_file = pd.read_csv(snakemake.input["concatenated_pvalues"], sep="\t", index_col=0)

for col in concatenated_pvalue_file.columns:
    new_col_name = col + " significant features"
    concatenated_pvalue_file.rename(columns={col: new_col_name}, inplace=True)


qadabra_all_result = pd.concat([concatenated_diff_file, concatenated_pvalue_file], axis=1)

for index, row in qadabra_all_result.iterrows():
    count = 0
    for col in qadabra_all_result.columns:
        if col.endswith(" significant features"):
            if row[col] < 0.05:
                count += 1
                qadabra_all_result.at[index, col] = "p<0.05"
            else:
                qadabra_all_result.at[index, col] = "ns"
    qadabra_all_result.at[index, "# of tools p > 0.05"] = count


qadabra_all_result.to_csv(snakemake.output[0], sep="\t", index=True)




import logging

import numpy as np
import pandas as pd
import seaborn as sns

from bokeh.plotting import figure, output_file, save
from bokeh.layouts import column, row
from bokeh.models import ColumnDataSource, Select, HoverTool
from bokeh.models.callbacks import CustomJS


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

output_file(filename=snakemake.output[0], title="Rank Comparison")

pval_df = pd.read_table(snakemake.input[0], sep="\t", index_col=0)
rank_df = (
    pval_df.rename(columns={f"{x}": f"{x}_pvalue" for x in pval_df.columns})
    .reset_index()
)

tool_list = pval_df.columns.tolist()
tool_rank_list = [x + "_pvalue" for x in tool_list]

# Rank Comparison
# logger.info("Creating rank comparison panel...")
chosen_tool_1 = Select(options=tool_rank_list, value=tool_rank_list[0], title="x-axis")
chosen_tool_2 = Select(options=tool_rank_list, value=tool_rank_list[1], title="y-axis")

rank_df["x"] = rank_df[chosen_tool_1.value]
rank_df["y"] = rank_df[chosen_tool_2.value]

source = ColumnDataSource(rank_df)

hover = HoverTool(mode="mouse", tooltips=[("Name", "points")], attachment="below")

hover.tooltips = (
    [("Feature ID", "@feature_id")] +
    [(f"{x}", f"@{x}") for x in rank_df.columns if x not in ["feature_id", "x", "y"]]
)

tools = [
    hover,
    "box_zoom",
    "reset",
    "pan",
    "save"
]

# plot = figure(tools=tools)
plot = figure(x_range=(0, 1), y_range=(0, 1), tools=tools)

plot.circle(
    source=source,
    x="x",
    y="y",
    color="lightblue",
    line_width=0.5,
    line_color="black",
    size=10,
    name="points"
)
plot.line(
    x=np.arange(0, rank_df.shape[0]),
    y=np.arange(0, rank_df.shape[0]),
    line_width=3,
    line_dash="dashed",
    color="black"
)

tool_callback = CustomJS(args=dict(source=source, plot=plot), code="""
const data = source.data[cb_obj.value];
if (cb_obj.title == 'x-axis') {
    plot.below[0].axis_label = cb_obj.value;
    source.data.x = data;
} else {
    plot.left[0].axis_label = cb_obj.value;
    source.data.y = data;
}

source.change.emit();
""")
chosen_tool_1.js_on_change("value", tool_callback)
chosen_tool_2.js_on_change("value", tool_callback)

plot.title.text = "P-value Comparison"
plot.title.text_font_size = "20pt"

plot.xaxis.axis_label = chosen_tool_1.value
plot.yaxis.axis_label = chosen_tool_2.value

for ax in [plot.xaxis, plot.yaxis]:
    ax.axis_label_text_font_size = "14pt"
    ax.axis_label_text_font_style = "normal"
    ax.major_label_text_font_size = "10pt"

layout = row(column(chosen_tool_1, chosen_tool_2), plot)
save(layout)

logger.info(f"Saved rank comparisons to {snakemake.output[0]}")



import logging

import pandas as pd


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logger.info("Loading differentials...")
diffs = pd.read_table(snakemake.input[0], sep="\t", index_col=0)
if snakemake.wildcards.get("tool") is not None:
    tool = snakemake.wildcards["tool"]
else:
    tool = "PC1"
pctile = int(snakemake.wildcards["pctile"])

sorted_diffs = diffs.sort_values(by=tool, ascending=False)

n = int(diffs.shape[0]*pctile/100)
top_feats = sorted_diffs.head(n).index.tolist()
bot_feats = sorted_diffs.tail(n).index.tolist()
top_feats = pd.DataFrame(
    dict(feature_id=top_feats, location="numerator")
)
bot_feats = pd.DataFrame(
    dict(feature_id=bot_feats, location="denominator")
)
df = pd.concat([top_feats, bot_feats])
df["pctile"] = pctile
df["num_feats"] = n

df.to_csv(snakemake.output[0], sep="\t", index=False)
logger.info(f"Saved to {snakemake.output[0]}")



import logging
import pandas as pd
from bokeh.plotting import output_file, save
from bokeh.models import DataTable, TableColumn, ColumnDataSource


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

output_file(filename=snakemake.output[0], title="Table")
diff_df = pd.read_table(snakemake.input[0], sep="\t", index_col=0)

diff_df = diff_df.round(2)
diff_df = diff_df.reset_index()

source = ColumnDataSource(diff_df)
columns = [
    TableColumn(field=x, title=x)
    for x in diff_df.columns
]
data_table = DataTable(source=source, columns=columns,
                       frozen_columns=1,
                       sizing_mode="stretch_height")
save(data_table)
logger.info(f"Saved table to {snakemake.output[0]}")


import logging

import biom
import numpy as np
import pandas as pd


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logger.info("Loading table...")
table = biom.load_table(snakemake.input["table"])
table = table.to_dataframe(dense=True).T

def log_ratio(table, top_feats, bot_feats):
    num_sum = table.loc[:, top_feats].sum(axis=1)
    denom_sum = table.loc[:, bot_feats].sum(axis=1)
    lr_df = pd.concat([num_sum, denom_sum], axis=1)
    lr_df.columns = ["num", "denom"]
    lr_df = lr_df.dropna(how="all")
    lr_df = lr_df + 1
    lr_df["log_ratio"] = np.log(lr_df["num"]/lr_df["denom"]).to_frame()
    return lr_df


feat_df = pd.read_table(snakemake.input["feats"], sep="\t")
if not feat_df.empty:
    top_feats = feat_df.query("location == 'numerator'")["feature_id"]
    bot_feats = feat_df.query("location == 'denominator'")["feature_id"]

    logger.info("Creating log-ratios...")
    lr_df = log_ratio(table, top_feats, bot_feats).reset_index()
    lr_df["pctile"] = snakemake.wildcards["pctile"]
    lr_df["num_feats"] = feat_df["num_feats"].unique().item()

    if snakemake.wildcards.get("tool") is not None:
        lr_df["tool"] = snakemake.wildcards["tool"]
    else:
        lr_df["tool"] = "pca_pc1"

    lr_df = lr_df.rename(columns={"index": "sample_name"})
    lr_df.to_csv(snakemake.output[0], sep="\t", index=False)
    logger.info(f"Saved to {snakemake.output[0]}")
else:
    logger.info("The feature DataFrame is empty. Skipping the rest of the script.")
    lr_df = pd.DataFrame()
    lr_df.to_csv(snakemake.output[0], sep="\t", index=False)
    logger.info(f"Saved to {snakemake.output[0]}")


import logging

from joblib import dump
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import (roc_curve, auc, precision_recall_curve,
                             average_precision_score)
from sklearn.model_selection import RepeatedStratifiedKFold


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

covariate = snakemake.params["factor_name"]
target = snakemake.params["target"]

logger.info(f"Covariate: {covariate}")
logger.info(f"Target: {target}")

metadata = pd.read_table(snakemake.input["metadata"], sep="\t",
                         index_col=0).dropna(subset=[covariate])
try:
    log_ratios = pd.read_table(snakemake.input["log_ratios"], sep="\t", index_col=0).join(metadata, how="inner")
    X = log_ratios["log_ratio"].values.reshape(-1, 1)
    y = (log_ratios[covariate] == target).astype(int).values

    mean_fpr = np.linspace(0, 1, 100)
    mean_rec = np.linspace(0, 1, 100)

    n_splits = snakemake.config["ml_params"]["n_splits"]
    n_repeats = snakemake.config["ml_params"]["n_repeats"]
    logger.info(
        f"Using repeated stratified k-fold CV with {n_splits} splits & "
        f"{n_repeats} repeats"
    )
    cv = RepeatedStratifiedKFold(n_splits=n_splits, n_repeats=n_repeats,
                                random_state=1)

    model = LogisticRegression(penalty="none")
    folds = [(train, test) for train, test in cv.split(X, y)]
    model_data = {
        "folds": {
            f"fold_{i+1}": {"train_idx": train, "test_idx": test}
            for i, (train, test) in enumerate(folds)
        }
    }
    model_data["fprs"] = mean_fpr
    model_data["recalls"] = mean_rec

    model_data["log_ratios"] = X
    model_data["truth"] = y

    # https://stackoverflow.com/a/67968945
    tprs = []
    precs = []
    roc_aucs = []
    pr_aucs = []
    for i, (train, test) in enumerate(folds):
        prediction = model.fit(X[train], y[train]).predict_proba(X[test])

        fpr, tpr, _ = roc_curve(y[test], prediction[:, 1])
        tpr_interp = np.interp(mean_fpr, fpr, tpr)
        tprs.append(tpr_interp)
        roc_auc = auc(mean_fpr, tpr_interp)
        roc_aucs.append(roc_auc)

        prec, rec, _ = precision_recall_curve(y[test], prediction[:, 1])
        prec = prec[::-1]
        rec = rec[::-1]
        prec_interp = np.interp(mean_rec, rec, prec)
        precs.append(prec_interp)
        pr_auc = auc(mean_rec, prec_interp)
        pr_aucs.append(pr_auc)

        logger.info(
            f"Fold {i+1}/{len(folds)}: ROC AUC = {roc_auc:.2f}, "
            f"PR AUC = {pr_auc:.2f}"
        )

    model_data["tprs"] = tprs
    model_data["precisions"] = precs
    model_data["roc_aucs"] = roc_aucs
    model_data["pr_aucs"] = pr_aucs

    mean_roc_auc = np.mean(roc_aucs)
    mean_pr_auc = np.mean(pr_aucs)

    logger.info(f"Mean ROC AUC = {mean_roc_auc:.2f}")
    logger.info(f"Mean PR AUC = {mean_pr_auc:.2f}")

    dump(model_data, snakemake.output[0], compress=True)
    logger.info(f"Saving model to {snakemake.output[0]}")
except:
    logger.error(f"The file {snakemake.input['log_ratios']} is empty or does not exist")
    model_data = {}
    dump(model_data, snakemake.output[0], compress=True)
    logger.info(f"Saving empty model to {snakemake.output[0]}")



import logging

import matplotlib.pyplot as plt
import pandas as pd
from upsetplot import UpSet, from_contents


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

tool_names = snakemake.config["tools"]
pctile = snakemake.wildcards["pctile"]
logger.info("Loading percentile features...")
feat_df_dict = {
    tool: pd.read_table(f, sep="\t", index_col=0)
    for tool, f in zip(tool_names, snakemake.input)
}

num_list = {
    tool: df.query("location == 'numerator'").index.tolist()
    for tool, df in feat_df_dict.items()
}

denom_list = {
    tool: df.query("location == 'denominator'").index.tolist()
    for tool, df in feat_df_dict.items()
}

try:
    num_plt = UpSet(from_contents(num_list), subset_size="count", show_counts=True).plot()
    plt.title(f"Numerator - {pctile}%")
except IndexError:
    plt.figure()
    plt.title(f"Numerator - {pctile}% (empty)")
    logger.info("num_list is empty. Creating empty numerator plot.")
finally:
    plt.savefig(snakemake.output["numerator"])
    logger.info(f"Saved to {snakemake.output['numerator']}")

try:
    denom_plt = UpSet(from_contents(denom_list), subset_size="count", show_counts=True).plot()
    plt.title(f"Denominator - {pctile}%")
except IndexError:
    plt.figure()
    plt.title(f"Denominator - {pctile}% (empty)")
    logger.info("denom_list is empty. Creating empty denominator plot.")
finally:
    plt.savefig(snakemake.output["denominator"])
    logger.info(f"Saved to {snakemake.output['denominator']}")



import logging
import pandas as pd

logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logger.info(f"Loading {snakemake.wildcards['tool']} p-values...")
col = snakemake.params["col"]
logger.info(f"Using '{col}' as column")
pvalues = pd.read_table(snakemake.input[0], sep="\t", index_col=0)[col]
pvalues.name = snakemake.wildcards["tool"]
pvalues.index.name = "feature_id"
pvalues.to_csv(snakemake.output[0], sep="\t", index=True)
logger.info(f"Saved to {snakemake.output[0]}")



import logging
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from bokeh.plotting import figure, save
from bokeh.models import ColumnDataSource, HoverTool
from bokeh.io import output_file

logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

logger.info("Loading differentials...")
diffs = pd.read_table(snakemake.input[0], sep="\t", index_col=0).squeeze()
values = diffs.sort_values().values

feature_names = diffs.sort_values().index

logger.info("Plotting differentials...")

# Create a ColumnDataSource to store the data
source = ColumnDataSource(data=dict(x=np.arange(len(feature_names)), feature_names=feature_names, values=values))

# Create the figure
# p = figure(x_range=(0, len(values)), height=400, width=600, toolbar_location=None)
p = figure(x_range=(0, len(values)), toolbar_location=None)

# Plot the bars
p.vbar(x='x', top='values', width=0.8, source=source)

# Add HoverTool to display information on hover
hover = HoverTool(tooltips=[("Feature Name", "@feature_names")])
p.add_tools(hover)

# Set plot properties
p.xaxis.axis_label = "Rank"
p.xgrid.grid_line_color = 'gray'
p.ygrid.grid_line_color = 'gray'

# Configure the title
tool_name = snakemake.wildcards["tool"]
p.title.text = f"{tool_name} Differentials"
p.title.align = 'center'
p.title.text_font_size = '18pt'

# Save plot
output_file(snakemake.output[0])
save(p)

logger.info(f"Saved interactive rank plot to {snakemake.output[0]}")

### the below code makes the non-interactive rank plots

# fig, ax = plt.subplots(1, 1)
# plt.bar(
#     np.arange(len(values)),
#     values,
#     width=1
# )
# tool_name = snakemake.wildcards["tool"]
# ax.set_title(f"{tool_name} Differentials")
# ax.set_xlabel("Rank")
# ax.grid("both")
# plt.savefig(snakemake.output[0])
# logger.info(f"Saved to {snakemake.output[0]}")


# logger.info("Plotting differentials...")
# fig, ax = plt.subplots(1, 1)
# plt.bar(
#     np.arange(len(values)),
#     values,
#     width=1
# )
# tool_name = snakemake.wildcards["tool"]
# ax.set_title(f"{tool_name} Differentials")
# ax.set_xlabel("Rank")
# ax.grid("both")
# plt.savefig(snakemake.output[0])
# logger.info(f"Saved to {snakemake.output[0]}")




import logging
import pathlib
from typing import List

import biom
import pandas as pd
import warnings

def _validate_input(
    logger: logging.Logger,
    table: pathlib.Path,
    metadata: pathlib.Path,
    factor_name: str,
    target_level: str,
    reference_level: str,
    tree: pathlib.Path = None,
    confounders: List[str] = None,
):
    logger.info("Loading metadata...")
    md = pd.read_table(metadata, sep="\t", index_col=0, dtype={0:'str'})

    if factor_name not in md.columns:
        raise ValueError(f"{factor_name} not found in metadata!")

    logger.info("Making sure factor & levels are all present in metadata...")
    factor_counts = md[factor_name].value_counts()
    logger.info(f"Factor counts: \n{factor_counts}")
    for level in [target_level, reference_level]:
        if level not in factor_counts:
            raise ValueError(f"{level} not found in {factor_name} values!")

    logger.info("Making sure confounders are all metadata columns...")
    for conf in confounders:
        if conf not in md.columns:
            raise ValueError(f"{conf} not found in metadata!")

    logger.info("Loading table...")
    tbl = biom.load_table(table)
    logger.info(f"Table shape: {tbl.shape}")
    tbl_idx = {str(item) for item in set(tbl.ids())}
    md_idx = {str(item) for item in set(md.index)}

    if not tbl_idx.issubset(md_idx):
        raise ValueError("Table IDs are not a subset of metadata IDs!")

    if tbl_idx != md_idx:
        logger.warn("Table IDs and metadata IDs are not exactly the same.")

    logger.info("Looking for completely discriminatory taxa...")
    tbl_df = tbl.to_dataframe(dense=True).T
    joint_df = tbl_df.join(md)
    gb = joint_df.groupby(factor_name).sum(numeric_only=True)
    feat_presence = gb.apply(lambda x: x.all())

    discriminating_feats = feat_presence[~feat_presence].index.tolist()

    if len(discriminating_feats) > 0:
        logger.warn("Number of discriminating features: " + str(len(discriminating_feats)))
        warning_msg = f"Some features in the table perfectly discriminate factor groups. Automatically filtering out {len(discriminating_feats)} features before running Qadabra..."        
        warnings.warn(warning_msg, category=Warning)

        # Filtering out the discriminating features from the BIOM table
        tbl = tbl.filter(lambda value, id_, metadata: id_ not in discriminating_feats, axis='observation', inplace=False)
        logger.info(f"Table shape after filtering: {tbl.shape}")


    if tree:
        from bp import parse_newick, to_skbio_treenode
        logger.info("Reading phylogenetic tree...")
        with open(tree) as f:
            tr = parse_newick(f.readline())
        tr = to_skbio_treenode(tr)
        tip_names = set(tip.name for tip in tr.tips())
        tbl_features = set(tbl.ids("observation"))
        if not tip_names.issubset(tbl_features):
            raise ValueError("Tree tips are not a subset of table features!")
    else:
        logger.info("Reading phylogenetic tree...")
        logger.info("(Optional tree file not provided. Skipping tree validation.)")


import logging

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logger.info("Loading differentials...")
diff_df = pd.read_table(snakemake.input[0], sep="\t", index_col=0)
rank_df = (
    diff_df.rank(ascending=False)
    .rename(columns={f"{x}": f"{x}_rank" for x in diff_df.columns})
    .join(diff_df)
)

tool_list = diff_df.columns.tolist()
tool_rank_list = [x + "_rank" for x in tool_list]

logger.info("Scaling values...")
scaled_diff_values = StandardScaler().fit_transform(diff_df.values)

pc_cols = [f"PC{x+1}" for x in range(len(tool_list))]
logger.info("Running PCA...")
pca = PCA(n_components=len(tool_list)).fit(scaled_diff_values)
feat_df = pd.DataFrame(
    pca.transform(scaled_diff_values),
    index=diff_df.index,
    columns=pc_cols
).join(rank_df)
max_vals = feat_df[pc_cols].max()

tool_df = pd.DataFrame(
    np.dot(pca.components_, np.diag(max_vals)),
    index=diff_df.columns,
    columns=pc_cols
)
tool_df.index.name = "tool"

feat_df.to_csv(snakemake.output["features"], sep="\t", index=True)
tool_df.to_csv(snakemake.output["tools"], sep="\t", index=True)

logger.info(f"Saved feature PCs to {snakemake.output['features']}")
logger.info(f"Saved tool PCs to {snakemake.output['tools']}")

prop_exp = pca.explained_variance_ratio_
prop_exp = pd.Series(prop_exp, index=pc_cols)
prop_exp.name = "proportion_var_explained"
prop_exp.to_csv(snakemake.output["prop_exp"], sep="\t", index=True)
logger.info(f"Saved explained variance to {snakemake.output['prop_exp']}")



# Much of this page is taken from the gemelli setup file
import ast
import re
from setuptools import find_packages, setup

# version parsing from __init__ pulled from Flask's setup.py
# https://github.com/mitsuhiko/flask/blob/master/setup.py
_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("qadabra/__init__.py", "rb") as f:
    hit = _version_re.search(f.read().decode("utf-8")).group(1)
    version = str(ast.literal_eval(hit))

with open("README.md") as f:
    long_description = f.read()

classes = """
    Development Status :: 4 - Beta
    License :: OSI Approved :: BSD License
    Topic :: Software Development :: Libraries
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Bio-Informatics
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3 :: Only
    Operating System :: Unix
    Operating System :: POSIX
    Operating System :: MacOS :: MacOS X
"""
classifiers = [s.strip() for s in classes.split("\n") if s]

description = "Differential abundance workflow for microbiome data"

standalone = ["qadabra=qadabra.qadabra:qadabra"]

setup(
    name="qadabra",
    author="Gibraan Rahman",
    author_email="grahman@eng.ucsd.edu",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gibsramen/qadabra",
    version=version,
    license="BSD-3-Clause",
    packages=find_packages(),
    install_requires=[
        "biom-format",
        "pandas>=1.0.0",
        "scikit-bio",
        "iow",
        "click",
    ],
    classifiers=classifiers,
    include_package_data=True,
    package_data={"qadabra": ["qadabra/workflow/*", "qadabra/config/*",
                              "qadabra/test_data/*"]},
    entry_points={"console_scripts": standalone}
)



import logging
import os
import pathlib
from pkg_resources import resource_filename
import shutil

import biom
import click
import pandas as pd

from qadabra import __version__
from qadabra.utils import _validate_input

SNKFILE_TEXT = """from pkg_resources import resource_filename

from snakemake.utils import min_version
min_version("6.0")

qadabra_snakefile = resource_filename("qadabra", "workflow/Snakefile")
configfile: "config/config.yaml"

module qadabra:
    snakefile:
        qadabra_snakefile
    config:
        config

use rule * from qadabra
"""


@click.group()
@click.version_option(__version__)
def qadabra():
    """Differential abundance workflow"""
    pass


@qadabra.command()
@click.option(
    "--workflow-dest",
    type=click.Path(exists=True),
    required=True,
    help="Location of workflow",
    default="."
)
@click.option(
    "--table",
    type=click.Path(exists=True),
    required=True,
    help="Feature table in BIOM format"
)
@click.option(
    "--metadata",
    type=click.Path(exists=True),
    required=True,
    help="Metadata in TSV format"
)
@click.option(
    "--tree",
    type=click.Path(exists=True),
    required=False,
    help="Phylogenetic tree in Newick format"
)
@click.option(
    "--name",
    type=str,
    required=True,
    help="Name of dataset"
)
@click.option(
    "--factor-name",
    type=str,
    required=True,
    help="Name of factor grouping in metadata"
)
@click.option(
    "--target-level",
    type=str,
    required=True,
    help="Grouping level on which to perform differential abundance"
)
@click.option(
    "--reference-level",
    type=str,
    required=True,
    help="Grouping level to use as reference"
)
@click.option(
    "--confounder",
    type=str,
    required=False,
    multiple=True,
    help="Confounder variable to consider (can provide multiple)"
)
@click.option(
    "--validate-input/--no-validate-input",
    show_default=True,
    default=True,
    help="Whether to validate input prior to adding dataset"
)
@click.option(
    "--verbose",
    is_flag=True,
    show_default=True,
    default=False,
    help="Whether to output progress to console"
)
def add_dataset(
    workflow_dest,
    table,
    metadata,
    tree,
    name,
    factor_name,
    target_level,
    reference_level,
    confounder,
    validate_input,
    verbose
):
    """Add dataset on which to run Qadabra"""
    wflow_path = pathlib.Path(workflow_dest)
    wflow_dir = wflow_path / "workflow"
    cfg_dir = wflow_path / "config"

    if not wflow_dir.exists:
        raise ValueError("Workflow has not been created!")

    if not cfg_dir.exists:
        raise ValueError("Config directory has not been created!")

    dataset_sheet = cfg_dir / "datasets.tsv"
    logger = logging.getLogger(__name__)
    log_level = logging.INFO if verbose else logging.WARNING
    logger.setLevel(log_level)
    sh = logging.StreamHandler()
    sh.setLevel(log_level)
    formatter = logging.Formatter(
        "[%(asctime)s - %(levelname)s] :: %(message)s",
        "%Y-%m-%d %H:%M:%S"
    )
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    if validate_input:
        logger.info("Validating input...")
        _validate_input(logger, table, metadata, factor_name, target_level,
                        reference_level, tree, confounder)

    dataset_sheet = pathlib.Path(dataset_sheet)
    new_ds = pd.Series({
        "table": pathlib.Path(table).resolve(),
        "metadata": pathlib.Path(metadata).resolve(),
        "factor_name": factor_name,
        "target_level": target_level,
        "reference_level": reference_level,
    }, name=name).to_frame().T

    if tree is not None:
        new_ds["tree"] = pathlib.Path(tree).resolve()
    else:
        new_ds["tree"] = None

    if confounder:
        new_ds["confounders"] = ";".join(confounder)
    else:
        new_ds["confounders"] = None

    if dataset_sheet.exists():
        logger.info("Loading datasheet...")
        ds_sheet = pd.read_table(dataset_sheet, sep="\t", index_col=0)
        if name in ds_sheet.index:
            raise ValueError(f"{name} already exists in dataset sheet!")
        ds_sheet = pd.concat([ds_sheet, new_ds], axis=0)
    else:
        logger.info("Dataset does not exist. Creating...")
        ds_sheet = new_ds

    ds_sheet.to_csv(dataset_sheet, sep="\t", index=True)
    logger.info(f"Saved dataset sheet to {dataset_sheet}")


@qadabra.command()
@click.option(
    "--workflow-dest",
    type=click.Path(exists=False),
    default="."
)
def create_workflow(workflow_dest):
    """Create new Qadabra workflow structure"""
    wflow_dest = pathlib.Path(workflow_dest)
    wflow_dir = wflow_dest / "workflow"
    cfg_dir = wflow_dest/ "config"
    os.makedirs(wflow_dir)
    os.makedirs(cfg_dir)

    cfg_file = resource_filename("qadabra", "config/config.yaml")
    shutil.copy(cfg_file, cfg_dir / "config.yaml")

    style_file = resource_filename("qadabra", "config/qadabra.mplstyle")
    shutil.copy(style_file, cfg_dir / "qadabra.mplstyle")

    snkfile_path = wflow_dir / "Snakefile"
    with open(snkfile_path, "w") as f:
        f.write(SNKFILE_TEXT)


if __name__ == "__main__":
    qadabra()



import logging

import matplotlib.pyplot as plt
from matplotlib.patches import Patch
import numpy as np
import pandas as pd
import seaborn as sns


logger = logging.getLogger("qadabra")
logger.setLevel(logging.INFO)
fh = logging.FileHandler(snakemake.log[0], mode="w")
formatter = logging.Formatter(
    f"[%(asctime)s - {snakemake.rule}] :: %(message)s"
)
fh.setFormatter(formatter)
logger.addHandler(fh)

logging.captureWarnings(True)
logging.getLogger("py.warnings").addHandler(fh)

plt.style.use(snakemake.config["stylesheet"])

logger.info("Loading PCA results...")
feat_df = pd.read_table(snakemake.input["features"], sep="\t", index_col=0)
tool_df = pd.read_table(snakemake.input["tools"], sep="\t", index_col=0)
prop_exp = pd.read_table(snakemake.input["prop_exp"], sep="\t", index_col=0)
prop_exp = prop_exp.squeeze()

pc_cols = tool_df.columns.tolist()
tool_list = tool_df.index.tolist()

arrow_palette = dict(zip(
    tool_list,
    sns.color_palette("colorblind", len(tool_list)).as_hex()
))

fig, ax = plt.subplots(1, 1)

logger.info("Plotting scatterplot...")
sns.scatterplot(
    data=feat_df,
    x="PC1",
    y="PC2",
    color="lightgray",
    edgecolor="black",
    linewidth=0.1,
    ax=ax
)

xlim = ax.get_xlim()
ylim = ax.get_ylim()

xrange = xlim[1] - xlim[0]
yrange = ylim[1] - ylim[0]

logger.info("Plotting arrows...")
for i, row in tool_df.iterrows():
    tool_name = row.name
    color = arrow_palette[tool_name]
    ax.arrow(
        x=0, y=0,
        dx=row["PC1"], dy=row["PC2"],
        facecolor=color,
        linewidth=0.5,
        head_width=xrange*0.005*5,
        width=xrange*0.005,
        edgecolor="black",
    )


patches = []
for i, row in tool_df.iterrows():
    tool_name = row.name
    color = arrow_palette[tool_name]
    ax.arrow(
        x=0, y=0,
        dx=row["PC1"], dy=row["PC2"],
        facecolor=color,
        linewidth=0.5,
        head_width=xrange*0.005*5,
        width=xrange*0.005,
        edgecolor="black",
    )
    tool_patch = Patch(
        facecolor=color,
        edgecolor="black",
        label=tool_name,
        linewidth=0.5,
    )
    patches.append(tool_patch)

ax.legend(
    handles=patches,
    bbox_to_anchor=[1, 1],
    loc="upper left",
    frameon=False
)

prop_exp_labels = [
    f"{i} ({x*100:.2f}%)"
    for i, x in prop_exp.items()
]
ax.set_xlabel(prop_exp_labels[0])
ax.set_ylabel(prop_exp_labels[1])

ax.set_title("PCA")
plt.savefig(snakemake.output[0])
logger.info(f"Saved to {snakemake.output[0]}")
